#
#  Copyright Bradley Lowekamp
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0.txt
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
""" Provides command line interfaces for sitkibex registration algorithms"""

import sitkibex.globals
import os
import sys
from os.path import basename
import SimpleITK as sitk
import click
import logging
import re
from .io import im_read_channel
import sitkibex.registration_utilities as utils


_logger = logging.getLogger(__name__)


class _Bunch(object):
    def __init__(self, adict):
        self.__dict__.update(adict)


class ImagePathChannel(click.Path):
    """
    A custom click ParamType.

    This class parses channel name suffixes for command line parameters.
    The '@CHANNEL` suffix can be appended to extract one channel. The `CHANNEL` can be a 0-based
    number e.g `@0`, or `Ch` followed by a 1-based index e.g. `@Ch1`, or it can be a string identifier
    for the name of a channel.

    If the channel suffix  is an integer ( or with  "Ch" ), then an integer type is returned for a 1-based channel
    index.


    The converted output returned is a pair of the filename followed by the channel name or integer channel index.

    """

    name = "image_file[@channel]"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def convert(self, value, param, ctx):

        m = re.match(r"(.*)@(.*)", value)

        channel_part = None
        if m:
            value = m.groups()[0]
            channel_part = m.groups()[1]

            m = re.match(r"(Ch)?(\d)", channel_part, re.IGNORECASE)
            if m:
                channel_part = int(m.groups()[1])
                # if "Ch" prefix it it 1 based index, so convert
                if m.groups()[0] is not None:
                    channel_part -= 1

        return super().convert(value, param, ctx), channel_part


@click.group()
@click.option(
    "--debug", "logging_level", flag_value=logging.DEBUG, help="Maximum verbosity with debugging logging enabled."
)
@click.option(
    "-v",
    "--verbose",
    "logging_level",
    flag_value=logging.INFO,
    default=True,
    help="Increased verbosity with information logging enabled.",
)
@click.option(
    "-q",
    "--quiet",
    "logging_level",
    flag_value=logging.WARNING,
    help="Minimal verbosity with only warning logging enabled.",
)
def cli(**kwargs):

    args = _Bunch(kwargs)

    # single app logger:
    log = sitkibex.globals.logger
    log.setLevel(args.logging_level)

    # Create handler to set everything at or below INFO to stdout
    h1 = logging.StreamHandler(sys.stdout)
    h1.setLevel(logging.NOTSET)
    h1.addFilter(lambda record: record.levelno <= logging.INFO)
    h1.setFormatter(logging.Formatter("%(message)s"))
    log.addHandler(h1)

    # Warnings and error go to stderr
    h2 = logging.StreamHandler(sys.stderr)
    h2.setLevel(logging.WARNING)
    h2.setFormatter(logging.Formatter("%(levelname)s - %(message)s"))
    log.addHandler(h2)

    if "SITK_SHOW_EXTENSION" not in os.environ:
        os.environ["SITK_SHOW_EXTENSION"] = ".nrrd"


@cli.command(name="registration")
@click.option(
    "-b",
    "--bin",
    default=1,
    type=int,
    show_default=True,
    help="Reduce the resolution of the input images in X and Y by this factor",
)
@click.option("-s", "--sigma", default=1.0, type=float, show_default=True)
@click.option("--affine/--no-affine", default=False, help="Do affine registration in a second step.")
@click.option(
    "--automask/--no-automask",
    default=False,
    show_default=True,
    help="Automatically compute a mask for the non-zero pixels of the input images",
)
@click.option(
    "--ignore-spacing/--no-ignore-spacing",
    default=True,
    show_default=True,
    help="Ignore the magnitude of spacing, but preserve relative ratio.",
)
@click.option(
    "--random/--no-random",
    default=False,
    show_default=True,
    help="Use wall-clock instead of a fixed seed for random initialization.",
)
@click.option("--samples-per-parameter", default=5000, type=int, show_default=True)
@click.argument("fixed_image", type=ImagePathChannel(exists=True, dir_okay=True, resolve_path=True))
@click.argument("moving_image", type=ImagePathChannel(exists=True, dir_okay=True, resolve_path=True))
@click.argument("output_transform", type=click.Path(exists=False, resolve_path=True))
def reg_cli(fixed_image, moving_image, output_transform, **kwargs):
    """Perform registration to solve for an OUTPUT_TRANSFORM mapping points from the FIXED_IMAGE to the MOVING_IMAGE."""
    from sitkibex.registration import registration

    args = _Bunch(kwargs)

    fixed_image, fixed_channel_name = fixed_image
    moving_image, moving_channel_name = moving_image

    if args.random:
        sitkibex.globals.default_random_seed = sitk.sitkWallClock

    def r(img, channel_name, bin_xy):
        img = im_read_channel(img, channel_name)
        img = sitk.Cast(img, sitk.sitkFloat32)
        if bin != 1:
            img = sitk.BinShrink(img, [bin_xy, bin_xy, 1])
        return img

    fixed_image = r(fixed_image, fixed_channel_name, args.bin)
    moving_image = r(moving_image, moving_channel_name, args.bin)

    tx = registration(
        fixed_image,
        moving_image,
        sigma=args.sigma,
        do_affine3d=args.affine,
        auto_mask=args.automask,
        ignore_spacing=args.ignore_spacing,
        samples_per_parameter=args.samples_per_parameter,
    )

    sitk.WriteTransform(tx, output_transform)


@cli.command(name="resample")
@click.option(
    "-b",
    "--bin",
    default=1,
    type=int,
    show_default=True,
    help="Reduce the resolution of the input images in X and Y by this factor",
)
@click.option("--fusion/--no-fusion", default=False, show_default=True, help="Blend fixed and moving images into RGB")
@click.option(
    "--combine/--no-combine",
    default=False,
    show_default=True,
    help="Combine fixed and moving images into multi-component",
)
@click.option("--invert/--no-invert", default=False, show_default=True, help="invert the transform")
@click.option(
    "--projection/--no-projection",
    default=False,
    show_default=True,
    help="project the volume in the z-direction to create 2-d image",
)
@click.option(
    "-o",
    "--output",
    default=None,
    help="filename for output image, if not provided the SimpleITK Show method is called",
    type=click.Path(exists=False, resolve_path=True),
)
@click.argument("fixed_image", type=ImagePathChannel(exists=True, dir_okay=True, resolve_path=True))
@click.argument("moving_image", type=ImagePathChannel(exists=True, dir_okay=True, resolve_path=True))
@click.argument("transform", required=False, type=click.Path(exists=True, dir_okay=False, resolve_path=True))
def resample_cli(fixed_image, moving_image, transform, **kwargs):
    """Create new image by transforming the MOVING_IMAGE onto the FIXED_IMAGE.

    Apply the TRANSFORM results from registration to resample the MOVING_IMAGE onto the coordinate space defined by the
    MOVING_IMAGE.

    To apply the results of registration:

    >>> sitkibex resample -o fixed_onto_moving.nrrd fixed.nrrd moving.nrrd out.txt

    Additional options can produce a small 2D RGB image to quickly review registration results. For example:

    >>> sitkibex resample --bin 10 --fusion --projection -o preview.png fixed.nrrd moving.nrrd out.txt

    If the moving image has more than 3 dimensions each sub-3d volume will be iteratively resampled.

    """

    from .resample import resample

    args = _Bunch(kwargs)

    @utils.sub_volume_execute()
    def binner(image):
        if args.bin != 1:
            return sitk.BinShrink(image, [args.bin, args.bin, 1])
        else:
            return image

    moving_image, moving_channel_name = moving_image
    fixed_image, fixed_channel_name = fixed_image

    moving_img = binner(im_read_channel(moving_image, moving_channel_name))

    if fixed_channel_name is None and os.path.isfile(fixed_image):
        reader = sitk.ImageFileReader()
        reader.SetFileName(fixed_image)
        reader.ReadImageInformation()
        if reader.GetDimension() > 3:
            if args.fusion:
                _logger.warning("Automatically selecting first channel with fusion enabled.")
            fixed_channel_name = 0

    fixed_img = im_read_channel(fixed_image, fixed_channel_name)
    fixed_img = binner(fixed_img)

    tx = None
    if transform:
        tx = sitk.ReadTransform(transform)

    @utils.sub_volume_execute(inplace=False)
    def resample_sub_volume(mv_img):
        return resample(
            fixed_img,
            mv_img,
            tx,
            fusion=args.fusion,
            combine=args.combine,
            invert=args.invert,
            projection=args.projection,
        )

    result = resample_sub_volume(moving_img)

    if args.output:
        sitk.WriteImage(result, args.output)
    else:
        sitk.Show(result, title="Resampling of {}".format(basename(fixed_image)))



#
#  Copyright Bradley Lowekamp
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0.txt
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
import SimpleITK as sitk

import sitkibex.registration_utilities as utils
import logging


logging.basicConfig(level=logging.DEBUG)


@utils.sub_volume_execute(inplace=True)
def do_add(img):
    img += 1.0
    return img


@utils.sub_volume_execute(inplace=False)
def do_add2(img):
    return img + 1.0


def test_sve_decorator_inplace_3d():
    img = sitk.Image([2, 3, 4], sitk.sitkUInt8)

    img_out = do_add(img)

    assert all([v == 1 for v in img_out])
    assert all([v == 1 for v in img])
    assert img.GetSize() == img_out.GetSize()


def test_sve_decorator_inplace_4d():
    img = sitk.Image([2, 3, 4, 5], sitk.sitkUInt8)

    img_out = do_add(img)

    assert all([v == 1 for v in img_out])
    assert all([v == 1 for v in img])
    assert img.GetSize() == img_out.GetSize()


def test_sve_decorator_inplace_5d():
    img = sitk.Image([2, 3, 4, 5, 6], sitk.sitkUInt8)

    img_out = do_add(img)

    assert all([v == 1 for v in img_out])
    assert all([v == 1 for v in img])
    assert img.GetSize() == img_out.GetSize()


def test_sve_decorator_3d():
    img = sitk.Image([2, 3, 4], sitk.sitkUInt8)

    img_out = do_add2(img)

    assert all([v == 1 for v in img_out])
    assert all([v == 0 for v in img])
    assert img.GetSize() == img_out.GetSize()


def test_sve_decorator_4d():
    img = sitk.Image([2, 3, 4, 5], sitk.sitkUInt8)

    img_out = do_add2(img)

    assert all([v == 1 for v in img_out])
    assert all([v == 0 for v in img])
    assert img.GetSize() == img_out.GetSize()


def test_sve_decorator_5d():
    img = sitk.Image([2, 3, 4, 5, 6], sitk.sitkUInt8)

    for c in range(img.GetSize()[4]):
        for t in range(img.GetSize()[3]):
            img[:, :, :, t, c] = 10 * t + c

    img_out = do_add2(img)

    assert all([v1 + 1 == v2 for v1, v2 in zip(img, img_out)])
    assert img.GetSize() == img_out.GetSize()



#
#  Copyright Bradley Lowekamp
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0.txt
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
import SimpleITK as sitk
import time
from functools import reduce
import itertools
from functools import wraps
import logging

_logger = logging.getLogger(__name__)


class RegistrationCallbackManager:
    """An object with member functions for the callbacks of a sitk RegistrationMethod"""

    def __init__(self, registration_method):
        self.R = registration_method
        self.start_time = None
        self.prev_time = None

    def add_command_callbacks(self, print_position=False, verbose=True):
        """Registers all callbacks with registration method"""

        if verbose:
            self.R.AddCommand(sitk.sitkIterationEvent, lambda: self.iteration_callback1(print_position=print_position))
        self.R.AddCommand(sitk.sitkEndEvent, lambda: self.end_callback())
        self.R.AddCommand(sitk.sitkStartEvent, lambda: self.start_callback())
        self.R.AddCommand(sitk.sitkMultiResolutionIterationEvent, lambda: self.multi_resolution_callback())

    def start_callback(self):
        _logger.info("===Registration Start===")
        self.start_time = time.time()

    def end_callback(self, message=None):
        end_time = time.time()
        if message is not None:
            _logger.info(message)

        _logger.info("Total elapsed time: {0:.4f}s".format(end_time - self.start_time))
        _logger.info("Optimizer stop condition: {0}".format(self.R.GetOptimizerStopConditionDescription()))
        _logger.info(" Iteration: {0}".format(self.R.GetOptimizerIteration()))
        _logger.info(" Final Metric value: {0}".format(self.R.GetMetricValue()))

        if self.R.GetInitialTransformInPlace():
            _logger.info(self.R.GetInitialTransform())

        _logger.info("===Registraion End===")

    def iteration_callback1(self, print_position=True):
        """
        A method to be added to a SimpleITK ImageRegistration method callback for the Iteration event.

        :param method: The SimpleITK.ImageRegistrationMethod object emitting the event.
        :param print_position: Disable printing the optimizer position or the transforms parameters.p
        """
        if self.R.GetOptimizerIteration() == 0 and len(self.R.GetOptimizerScales()):
            if self.R.GetOptimizerLearningRate() != 0:
                _logger.info("Learning Rate: {0}".format(self.R.GetOptimizerLearningRate()))

            if self.R.GetInitialTransform().IsLinear():
                _logger.info("Estimated Scales: {0}".format(self.R.GetOptimizerScales()))
            else:
                pass
                # _logger.info("Estimated Scales: {0}".format(self.R.GetOptimizerScales()[0]))

        number_of_valid_points = "?"
        try:
            number_of_valid_points = self.R.GetMetricNumberOfValidPoints()
        except AttributeError:
            # Ignore self.R not existing
            pass

        if print_position:
            _logger.info(
                "{0:3} = {1:10.5f} : {2} ({3:10.5e}) [#{4}]".format(
                    self.R.GetOptimizerIteration(),
                    self.R.GetMetricValue(),
                    self.R.GetOptimizerPosition(),
                    self.R.GetOptimizerConvergenceValue(),
                    number_of_valid_points,
                )
            )

        else:
            _logger.info(
                "{0:3} = {1:10.5f} ({2:10.5e}) [#{3}]".format(
                    self.R.GetOptimizerIteration(),
                    self.R.GetMetricValue(),
                    self.R.GetOptimizerConvergenceValue(),
                    number_of_valid_points,
                )
            )

    def multi_resolution_callback(self, message=None):

        multi_time = time.time()
        if self.R.GetCurrentLevel() > 0:
            _logger.info("Elapsed time: {0:.4f}s".format(multi_time - self.prev_time))
            _logger.info("Optimizer stop condition: {0}".format(self.R.GetOptimizerStopConditionDescription()))
            _logger.info(" Total Iteration: {0}".format(self.R.GetOptimizerIteration()))
            _logger.info(" Metric value: {0}".format(self.R.GetMetricValue()))
            _logger.info(" Convergence value: {0}".format(self.R.GetOptimizerConvergenceValue()))

        self.prev_time = multi_time
        _logger.info("---Level {0} Start---".format(self.R.GetCurrentLevel()))


def sub_volume_execute(inplace=True):
    """
    A function decorator which executes func on each sub-volume and *in-place* pastes the output input the
    input image.

    :param inplace:
    :param func: A function which take a SimpleITK Image as it's first argument and returns the results.
    :return: A wrapped function.
    """

    def wrapper(func):
        @wraps(func)
        def slice_by_slice(image, *args, **kwargs):

            dim = image.GetDimension()

            if dim <= 3:
                image = func(image, *args, **kwargs)
                return image

            extract_size = list(image.GetSize())
            extract_size[3:] = itertools.repeat(0, dim - 3)

            extract_index = [0] * dim
            paste_idx = [slice(None, None)] * dim

            extractor = sitk.ExtractImageFilter()
            extractor.SetSize(extract_size)
            if inplace:
                for high_idx in itertools.product(*[range(s) for s in image.GetSize()[3:]]):
                    extract_index[3:] = high_idx
                    extractor.SetIndex(extract_index)

                    paste_idx[3:] = high_idx
                    image[paste_idx] = func(extractor.Execute(image), *args, **kwargs)

            else:
                img_list = []
                for high_idx in itertools.product(*[range(s) for s in image.GetSize()[3:]]):
                    extract_index[3:] = high_idx
                    extractor.SetIndex(extract_index)

                    paste_idx[3:] = high_idx

                    img_list.append(func(extractor.Execute(image), *args, **kwargs))

                for d in range(3, dim):
                    step = reduce((lambda x, y: x * y), image.GetSize()[d + 1 :], 1)
                    img_list = [sitk.JoinSeries(img_list[i::step], image.GetSpacing()[d]) for i in range(step)]

                assert len(img_list) == 1
                image = img_list[0]

            return image

        return slice_by_slice

    return wrapper



#
#  Copyright Bradley Lowekamp
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0.txt
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from unittest import TestCase

import SimpleITK as sitk
from sitkibex import registration
import logging
import math

logging.basicConfig(level=logging.DEBUG)


class TestReg(TestCase):
    @staticmethod
    def generate_blob(point, size):
        img = sitk.GaussianSource(
            mean=point, sigma=[s / 10.0 for s in size], scale=255, size=size, outputPixelType=sitk.sitkFloat32
        )

        # img.SetSpacing(2.1e-6, 2.1e-6, 1.1e-5)

        return img

    @staticmethod
    def generate_double_blobs(point1, point2, size):
        img2 = __class__.generate_blob(point1, size)
        img1 = __class__.generate_blob(point2, size)
        return img1 + img2

    def check_near(self, pt1, pt2, tolerance=1e-4):

        self.assertEqual(len(pt1), len(pt2))
        distance = math.sqrt(sum([(float(i) - float(j)) ** 2 for i, j in zip(pt1, pt2)]))

        fail_msg = "The distance between {0} and {1} exceeded tolerance".format(pt1, pt2)
        self.assertLess(distance, tolerance, msg=fail_msg)

    def test_reg0(self):
        pts = [[256, 256, 8]]
        # all registration options set to False, so no registration is
        # performed, returning the identity transformation.
        tx = registration(
            sitk.Image([32, 32], sitk.sitkUInt8),
            sitk.Image([32, 32], sitk.sitkUInt8),
            do_fft_initialization=False,
            do_affine2d=False,
            do_affine3d=False,
        )
        self.check_near(pts[0], tx.TransformPoint(pts[0]))

    def test_reg1(self):
        fixed_pts = [[256, 256, 8], [64, 64, 7]]
        fixed = self.generate_double_blobs(point1=fixed_pts[0], point2=fixed_pts[1], size=[512, 511, 16])

        moving_pts = [[240, 288, 8], [48, 96, 7]]
        moving = self.generate_double_blobs(point1=moving_pts[0], point2=moving_pts[1], size=[512, 511, 16])

        tx = registration(fixed, moving, do_affine3d=False, do_fft_initialization=True)

        self.check_near(moving_pts[0], tx.TransformPoint(fixed_pts[0]))
        self.check_near(moving_pts[1], tx.TransformPoint(fixed_pts[1]))

    def test_reg2(self):
        fixed = self.generate_double_blobs(point1=[256, 256, 8], point2=[64, 64, 7], size=[512, 511, 16])

        moving = self.generate_double_blobs(point1=[240, 288, 8], point2=[48, 96, 7], size=[512, 511, 16])

        registration(fixed, moving, do_affine3d=True, do_fft_initialization=False, samples_per_parameter=100)

    def test_reg3(self):
        fixed = self.generate_double_blobs(point1=[256, 256, 8], point2=[64, 64, 7], size=[512, 511, 16])

        moving = self.generate_double_blobs(point1=[240, 288, 8], point2=[48, 96, 7], size=[512, 511, 16])

        registration(
            fixed, moving, do_affine3d=False, do_affine2d=True, do_fft_initialization=False, samples_per_parameter=100
        )

    def test_reg2_with_fft(self):
        fixed = self.generate_double_blobs(point1=[256, 256, 8], point2=[64, 64, 7], size=[512, 511, 16])

        moving = self.generate_double_blobs(point1=[240, 288, 8], point2=[48, 96, 7], size=[512, 511, 16])

        registration(fixed, moving, do_affine3d=True, do_fft_initialization=True, samples_per_parameter=100)

    def test_reg3_with_fft(self):
        fixed = self.generate_double_blobs(point1=[256, 256, 8], point2=[64, 64, 7], size=[512, 511, 16])

        moving = self.generate_double_blobs(point1=[240, 288, 8], point2=[48, 96, 7], size=[512, 511, 16])

        registration(
            fixed, moving, do_affine3d=False, do_affine2d=True, do_fft_initialization=True, samples_per_parameter=100
        )

    def test_reg4_with_fft(self):
        fixed = self.generate_double_blobs(point1=[256, 256, 8], point2=[64, 64, 7], size=[512, 511, 16])

        moving = self.generate_double_blobs(point1=[240, 288, 8], point2=[48, 96, 7], size=[512, 511, 16])

        registration(
            fixed, moving, do_affine3d=True, do_affine2d=True, do_fft_initialization=True, samples_per_parameter=100
        )

    def test_reg5_with_mask(self):
        fixed = self.generate_double_blobs(point1=[256, 256, 8], point2=[64, 64, 7], size=[512, 511, 16])
        fixed *= sitk.Cast(fixed >= 0.1, sitk.sitkFloat32)

        moving = self.generate_double_blobs(point1=[240, 288, 8], point2=[48, 96, 7], size=[512, 511, 16])

        registration(
            fixed,
            moving,
            do_affine3d=True,
            do_affine2d=False,
            do_fft_initialization=True,
            auto_mask=True,
            samples_per_parameter=500,
        )



#
#  Copyright Bradley Lowekamp
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0.txt
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import SimpleITK as sitk
import numpy as np
from .registration_utilities import RegistrationCallbackManager
from . import image_utilities as imgf
import sitkibex.globals
import logging

_logger = logging.getLogger(__name__)


def register_3d(
    fixed_image,
    moving_image,
    initial_transform,
    sigma_base=1.0,
    fixed_image_mask=None,
    moving_image_mask=None,
    number_of_samples_per_parameter=5000,
):
    """Perform multi-resolution 3D registration, with parameters tuned for affine transformation.


    :param fixed_image: a 3D SimpleITK image
    :param moving_image: a 3D SimpleITK image
    :param initial_transform: a SimpleITK transform initialized with center of transform and/or the transformation
     from a previous stage.
    :param sigma_base: scalar to change the amount of Gaussian smoothing performed
    :param fixed_image_mask: (optional) a binary image of non-zeros for pixels to use
    :param moving_image_mask: (optional) a binary image of non-zeros for pixels to use
    :param number_of_samples_per_parameter: Number of sample points per number of transform parameter to use for metric
     evaluation.
    :return:
    """
    use_neighborhood_correlation = False
    scale_factors = [4, 2, 1]

    reg = sitk.ImageRegistrationMethod()

    reg.MetricUseMovingImageGradientFilterOff()
    reg.MetricUseFixedImageGradientFilterOff()

    reg.SetOptimizerAsGradientDescentLineSearch(
        learningRate=1.0,
        numberOfIterations=400,
        convergenceMinimumValue=1e-7,
        convergenceWindowSize=10,
        lineSearchLowerLimit=0,
        lineSearchUpperLimit=1.0,
        lineSearchMaximumIterations=5,
        maximumStepSizeInPhysicalUnits=1.0,
    )

    reg.SetOptimizerScalesFromIndexShift()

    sampling_percentage = (
        initial_transform.GetNumberOfParameters() * number_of_samples_per_parameter / fixed_image.GetNumberOfPixels()
    )

    if fixed_image_mask or moving_image_mask:
        mask = fixed_image_mask
        if moving_image_mask:
            mask = moving_image_mask
        stats = sitk.StatisticsImageFilter()
        stats.Execute(mask != 0)

        _logger.info("Unmasked sampling: {0}".format(sampling_percentage))
        _logger.info("\tmasked percentage: {0}%".format(100 * stats.GetMean()))
        # this scaling assumes that the size in pixels of the mask the the fixed image are about the same
        sampling_percentage /= stats.GetMean()
        # sampling_percentage *= 10

        _logger.info("Adjusted sampling: {0}".format(sampling_percentage))

    if use_neighborhood_correlation:
        neighborhood_radius = 4
        reg.SetMetricAsANTSNeighborhoodCorrelation(radius=neighborhood_radius)
        sampling_percentage *= 0.1
    else:
        reg.SetMetricAsCorrelation()

    sampling_percentage_per_level = [min(0.10, sampling_percentage * f * f) for f in scale_factors]
    _logger.info(
        "Sampling Percentage Per Level: {0} #{1}".format(
            sampling_percentage_per_level,
            [
                p * fixed_image.GetNumberOfPixels() / (f**3)
                for p, f in zip(sampling_percentage_per_level, scale_factors)
            ],
        )
    )

    reg.SetMetricSamplingPercentagePerLevel(sampling_percentage_per_level, sitkibex.globals.default_random_seed)
    reg.SetMetricSamplingStrategy(reg.REGULAR)
    reg.SetShrinkFactorsPerLevel([f for f in scale_factors])
    reg.SmoothingSigmasAreSpecifiedInPhysicalUnitsOn()
    smoothing_sigmas = [2.0 * sigma_base * f * fixed_image.GetSpacing()[0] for f in scale_factors]
    reg.SetSmoothingSigmasPerLevel(smoothing_sigmas)
    _logger.info("SmoothingSigmasPerLevel: {}".format(smoothing_sigmas))

    reg.SetInterpolator(sitk.sitkLinear)

    if moving_image_mask:
        reg.SetMetricMovingMask(moving_image_mask)

    if fixed_image_mask:
        reg.SetMetricFixedMask(fixed_image_mask)

    reg.SetInitialTransform(initial_transform)

    reg_callbacks = RegistrationCallbackManager(reg)
    reg_callbacks.add_command_callbacks(print_position=True)

    return reg.Execute(fixed_image, moving_image)


def register_as_2d_affine(
    fixed_image,
    moving_image,
    sigma_base=1.0,
    initial_translation=None,
    fixed_image_mask=None,
    moving_image_mask=None,
    number_of_samples_per_parameter=5000,
):
    """Perform 2D registration from 3D image projected in the z-direction.

    The fixed_image and moving_image are first projected along the z-dimension to generate 2D images. Then a
    transform to map points from the fixed_image to the moving_image is optimized for normalized
    correlation metric. First a Euler2D transform is optimized, then a 2D Affine.

    :param fixed_image: a 3D SimpleITK Image class
    :param moving_image: a 3D SimpelITK Image class
    :param sigma_base: scalar to change the amount of Gaussian smoothing performed
    :param initial_translation: (optional) initial translation for fixed points to map to moving.
    :param fixed_image_mask: (optional) a binary image of non-zeros for pixels to use
    :param moving_image_mask: (optional) a binary image of non-zeros for pixels to use
    :param number_of_samples_per_parameter: Number of sample points per number of transform parameter to use for metric
     evaluation.
    :return: a 3D SimpleITK AffineTransform mapping points from the fixed_image to the moving_image
    """

    do_affine = True
    verbose = True

    _logger.info("Initializing projected registration...")
    _logger.info("Sigma Base: {0}".format(sigma_base))

    fixed_2d = imgf.project(fixed_image)
    moving_2d = imgf.project(moving_image)

    if fixed_image_mask:
        fixed_mask_2d = imgf.project(fixed_image_mask, projection_func=sitk.MedianProjection)
    else:
        fixed_mask_2d = None

    if moving_image_mask:
        moving_mask_2d = imgf.project(moving_image_mask, projection_func=sitk.MedianProjection)
    else:
        moving_mask_2d = None

    # Initialize the center of transform and align the center of the volumes.
    # - Use Euler transform
    #

    initial_rigid = sitk.CenteredTransformInitializer(
        fixed_2d, moving_2d, sitk.Euler2DTransform(), sitk.CenteredTransformInitializerFilter.GEOMETRY
    )

    if initial_translation:
        initial_rigid = sitk.Euler2DTransform(initial_rigid)
        initial_rigid.SetTranslation(initial_translation)

    #
    # Setup Multi-scale 2D Multi-scale registration
    #
    R = sitk.ImageRegistrationMethod()

    R.SetMetricAsCorrelation()

    # Due to the sparse sampling the gradient on the whole image
    # is not needed, and it is more time efficient not to compute.
    R.MetricUseMovingImageGradientFilterOff()
    R.MetricUseFixedImageGradientFilterOff()

    R.SetOptimizerAsGradientDescent(
        learningRate=1.0,
        numberOfIterations=500,
        convergenceMinimumValue=1e-6,
        convergenceWindowSize=10,
        maximumStepSizeInPhysicalUnits=2.0,
    )

    R.SetOptimizerScalesFromIndexShift()

    scale_factors = [16, 8, 4]
    # We don't need more samples for larger image, so base the number of samples on the number of parameters
    sampling_percentage = (
        len(initial_rigid.GetParameters()) * number_of_samples_per_parameter / fixed_2d.GetNumberOfPixels()
    )
    R.SetMetricSamplingPercentagePerLevel(
        [min(0.10, sampling_percentage) for f in scale_factors], sitkibex.globals.default_random_seed
    )
    R.SetMetricSamplingStrategy(R.REGULAR)
    R.SetShrinkFactorsPerLevel([1 for f in scale_factors])
    R.SmoothingSigmasAreSpecifiedInPhysicalUnitsOn()
    R.SetSmoothingSigmasPerLevel([4.0 * sigma_base * f * fixed_2d.GetSpacing()[0] for f in scale_factors])

    R.SetInitialTransform(initial_rigid)
    R.SetInterpolator(sitk.sitkLinear)

    if fixed_mask_2d:
        _logger.info("Setting fixed mask")
        R.SetMetricFixedMask(fixed_mask_2d)

    if moving_mask_2d:
        _logger.info("Setting moving mask")
        R.SetMetricMovingMask(moving_mask_2d)

    R_callbacks = RegistrationCallbackManager(R)
    R_callbacks.add_command_callbacks(print_position=True, verbose=verbose)

    # R.DebugOn()

    rigid_result_2d = R.Execute(fixed_2d, moving_2d)

    # Perform a deep copy of the transform to convert a sitk.Transform to the actual Euler type
    rigid_result_2d = sitk.Euler2DTransform(rigid_result_2d)

    # Promote Euler to 2D Affine
    affine = sitk.AffineTransform(2)
    affine.SetMatrix(rigid_result_2d.GetMatrix())
    affine.SetTranslation(rigid_result_2d.GetTranslation())
    affine.SetCenter(rigid_result_2d.GetCenter())

    #
    # Setup Multi-Scale 2D Affine registration
    #
    R2 = sitk.ImageRegistrationMethod()
    R2.SetMetricAsCorrelation()
    R2.MetricUseMovingImageGradientFilterOff()
    R2.MetricUseFixedImageGradientFilterOff()
    R2.SetOptimizerAsGradientDescentLineSearch(
        learningRate=1.0,
        numberOfIterations=100,
        convergenceMinimumValue=1e-6,
        convergenceWindowSize=10,
        lineSearchLowerLimit=0,
        lineSearchUpperLimit=2.0,
        lineSearchMaximumIterations=5,
        maximumStepSizeInPhysicalUnits=2,
    )

    R2.SetOptimizerScalesFromIndexShift()

    scale_factors = [8, 4, 2]
    sampling_percentage = len(affine.GetParameters()) * number_of_samples_per_parameter / fixed_2d.GetNumberOfPixels()
    R2.SetMetricSamplingPercentagePerLevel(
        [min(0.10, sampling_percentage * f) for f in scale_factors], sitkibex.globals.default_random_seed
    )
    R2.SetMetricSamplingStrategy(R.RANDOM)
    R2.SetShrinkFactorsPerLevel([f for f in scale_factors])
    R2.SmoothingSigmasAreSpecifiedInPhysicalUnitsOn()
    R2.SetSmoothingSigmasPerLevel([sigma_base * f * fixed_2d.GetSpacing()[0] for f in scale_factors])

    R2.SetInitialTransform(affine)
    R2.SetInterpolator(sitk.sitkLinear)

    R2_callbacks = RegistrationCallbackManager(R2)
    R2_callbacks.add_command_callbacks(print_position=True, verbose=verbose)

    if do_affine:
        affine_result = R2.Execute(fixed_2d, moving_2d)
    else:
        affine_result = affine

    # Do explicit casting
    affine_result = sitk.AffineTransform(affine_result)

    #
    # Convert 2d matrix into 3d
    #
    matrix_3d = np.identity(3)
    matrix_3d[:2, :2] = np.asarray(affine_result.GetMatrix()).reshape(2, 2)

    idx_2d = moving_2d.TransformPhysicalPointToContinuousIndex(affine_result.GetCenter())
    idx_3d = idx_2d + (moving_image.GetSize()[2] / 2.0,)
    center_3d = moving_image.TransformContinuousIndexToPhysicalPoint(idx_3d)

    # The 2d projected images preserve the image spacing but have the direction matrix set to the identity.
    _logger.info("center 2d->3d: {0}->{1}".format(affine_result.GetCenter(), center_3d))

    result_3d = sitk.AffineTransform(3)
    result_3d.SetTranslation(affine_result.GetTranslation() + (0,))
    result_3d.SetMatrix(matrix_3d.flatten())
    result_3d.SetCenter(center_3d)

    return result_3d


def registration(
    fixed_image: sitk.Image,  # noqa: C901
    moving_image: sitk.Image,
    *,
    do_fft_initialization=True,
    do_affine2d=False,
    do_affine3d=True,
    ignore_spacing=True,
    sigma=1.0,
    auto_mask=False,
    samples_per_parameter=5000,
    expand=None
) -> sitk.Transform:
    """Robust multi-phase registration for multi-panel confocal microscopy images.

    The fixed and moving image are expected to be the same molecular labeling, and the same imaged regioned.

    The phase available are:
      - fft initialization for translation estimation
      - 2D affine which can correct rotational acquisition problems, this phase is done on z-projections to optimize a \
       2D similarity transform followed by 2D affine
      - 3D affine robust mulit-level registration


    :param fixed_image: a scalar SimpleITK 3D Image
    :param moving_image: a scalar SimpleITK 3D Image
    :param do_fft_initialization: perform FFT based cross correlation for initialize translation
    :param do_affine2d: perform registration on 2D images from z-projection
    :param do_affine3d: multi-level affine transform
    :param ignore_spacing: internally adjust spacing magnitude to be near 1 to avoid numeric stability issues with \
    micro sized spacing
    :param sigma: scalar to change the amount of Gaussian smoothing performed
    :param auto_mask: ignore zero valued pixels connected to the image boarder
    :param samples_per_parameter: the number of image samples to used per transform parameter at full resolution
    :param expand: Perform super-sampling to increase number of z-slices by an integer factor. Super-sampling is \
    automatically performed when the number of z-slices is less than 5.
    :return: A SimpleITK transform mapping points from the fixed image to the moving. This may be a CompositeTransform.

    """
    # Identity transform will be returned if all registration steps are disabled by
    # the calling function.
    result = sitk.Transform()

    initial_translation_3d = True

    moving_mask = None
    fixed_mask = None

    number_of_samples_per_parameter = samples_per_parameter

    expand_factors = None

    if expand:
        expand_factors = [1, 1, expand]

    if fixed_image.GetPixelID() != sitk.sitkFloat32:
        fixed_image = sitk.Cast(fixed_image, sitk.sitkFloat32)

    # expand the image if at least 5 in any dimension
    if not expand_factors:
        expand_factors = [-(-5 // s) for s in fixed_image.GetSize()]

    if any([e != 1 for e in expand_factors]):
        _logger.warning(
            "Fixed image under sized in at lease one dimension!"
            "\tApplying expand factors {0} to image size.".format(expand_factors)
        )
        fixed_image = sitk.Expand(fixed_image, expandFactors=expand_factors)

    if moving_image.GetPixelID() != sitk.sitkFloat32:
        moving_image = sitk.Cast(moving_image, sitk.sitkFloat32)

    expand_factors = [-(-5 // s) for s in moving_image.GetSize()]
    if any([e != 1 for e in expand_factors]):
        _logger.warning(
            "WARNING: Moving image under sized in at lease one dimension!"
            "\tApplying expand factors {0} to image size.".format(expand_factors)
        )
        moving_image = sitk.Expand(moving_image, expandFactors=expand_factors)

    if auto_mask:
        fixed_mask = imgf.make_auto_mask(fixed_image)
        moving_mask = imgf.make_auto_mask(moving_image)

    if ignore_spacing:

        #
        # FORCE THE SPACING magnitude to be normalized near 1.0
        #

        spacing_magnitude = imgf.spacing_average_magnitude(fixed_image)

        _logger.info("Adjusting image spacing by {0}...".format(1.0 / spacing_magnitude))

        new_spacing = [s / spacing_magnitude for s in fixed_image.GetSpacing()]
        _logger.info("\tFixed Image Spacing: {0}->{1}".format(fixed_image.GetSpacing(), new_spacing))
        fixed_image.SetSpacing(new_spacing)
        fixed_image.SetOrigin([o / spacing_magnitude for o in fixed_image.GetOrigin()])

        new_spacing = [s / spacing_magnitude for s in moving_image.GetSpacing()]
        _logger.info("\tMoving Image Spacing: {0}->{1}".format(moving_image.GetSpacing(), new_spacing))
        moving_image.SetSpacing(new_spacing)
        moving_image.SetOrigin([o / spacing_magnitude for o in moving_image.GetOrigin()])

        if moving_mask:
            moving_mask.SetSpacing(new_spacing)
            moving_mask.SetOrigin([o / spacing_magnitude for o in moving_mask.GetOrigin()])

        if fixed_mask:
            fixed_mask.SetSpacing(new_spacing)
            fixed_mask.SetOrigin([o / spacing_magnitude for o in fixed_mask.GetOrigin()])

    #
    #
    # Do FFT based translation initialization
    #
    #
    initial_translation = None
    if do_fft_initialization:
        initial_translation = imgf.fft_initialization(
            moving_image, fixed_image, bin_shrink=8, projection=(not initial_translation_3d)
        )
        result = sitk.TranslationTransform(len(initial_translation), initial_translation)

    #
    # Do 2D registration first
    #
    if do_affine2d:
        result = register_as_2d_affine(
            fixed_image,
            moving_image,
            sigma_base=sigma,
            initial_translation=initial_translation,
            fixed_image_mask=fixed_mask,
            moving_image_mask=moving_mask,
        )

    if do_affine3d:

        _logger.info("Initializing Affine Registration...")

        if do_affine2d:
            # set the FFT xcoor initial z translation
            if do_fft_initialization and len(initial_translation) >= 3:

                # take the z-translation from the FFT
                translation = list(result.GetTranslation())
                translation[2] = initial_translation[2]
                result.SetTranslation(translation)

                _logger.info("Initialized 3D affine with z-translation... {0}".format(translation))

            affine = result
        else:
            affine = sitk.CenteredTransformInitializer(
                fixed_image, moving_image, sitk.AffineTransform(3), sitk.CenteredTransformInitializerFilter.GEOMETRY
            )
            affine = sitk.AffineTransform(affine)

            if do_fft_initialization:
                if len(initial_translation) >= 3:
                    affine.SetTranslation(list(initial_translation))
                    _logger.info("Initialized 3D affine with z-translation... {0}".format(initial_translation))
                else:
                    affine.SetTranslation(
                        list(initial_translation)
                        + [
                            0,
                        ]
                    )

        affine_result = register_3d(
            fixed_image,
            moving_image,
            initial_transform=affine,
            sigma_base=sigma,
            fixed_image_mask=fixed_mask,
            moving_image_mask=moving_mask,
            number_of_samples_per_parameter=number_of_samples_per_parameter,
        )

        result = affine_result

    if ignore_spacing:

        # Compose the scaling Transform into a single affine transform

        # The spacing of the image was modified to do registration, so we need to apply the appropriate scaling to
        # transform to the space the registration was done in.r

        scale = spacing_magnitude

        scale_transform = sitk.ScaleTransform(3)
        scale_transform.SetScale([scale] * 3)

        result = sitk.CompositeTransform([sitk.Transform(scale_transform), result, scale_transform.GetInverse()])

        # if result was a composite transform then we have nested composite
        # transforms white need to be flattened for writing.

        result.FlattenTransform()

        _logger.info(result)

    return result



#
#  Copyright Bradley Lowekamp
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0.txt
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from .registration import registration
from .resample import resample

try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown version"

__author__ = ["Bradley Lowekamp"]

__all__ = ["registration", "resample"]



#
#  Copyright Bradley Lowekamp
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0.txt
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import SimpleITK as sitk

import os.path
import logging
from .xml_info import XMLInfo, OMEInfo
from pathlib import Path
import numpy as np

try:
    import zarr

    _has_zarr = True
except ImportError:
    _has_zarr = False


_logger = logging.getLogger(__name__)

_vector_to_scalar = {
    sitk.sitkVectorUInt8: sitk.sitkUInt8,
    sitk.sitkVectorInt8: sitk.sitkInt8,
    sitk.sitkVectorUInt16: sitk.sitkUInt16,
    sitk.sitkVectorInt16: sitk.sitkInt16,
    sitk.sitkVectorUInt32: sitk.sitkUInt32,
    sitk.sitkVectorInt32: sitk.sitkInt32,
    sitk.sitkVectorFloat32: sitk.sitkFloat32,
    sitk.sitkVectorFloat64: sitk.sitkFloat64,
}


def _zarr_read_channel(filename: Path, channel=None) -> sitk.Image:
    """
    Read a channel from a zarr file.
    """

    store = zarr.DirectoryStore(filename)
    zarr_group = zarr.open_group(store=store, mode="r")

    axes = zarr_group.attrs["multiscales"][0]["axes"]
    ds_attr = zarr_group.attrs["multiscales"][0]["datasets"][0]

    arr = zarr_group[ds_attr["path"]]
    _logger.debug(arr)

    spacing = ds_attr["coordinateTransformations"][0]["scale"]

    axes_names = [ax["name"].upper() for ax in axes]

    if axes_names != list("TCZYX"):
        raise ValueError(f"Only TCZYX axes are supported, not {axes_names}.")

    if arr.shape[0] > 1:
        raise ValueError("Only single time point is supported.")

    if channel is None:
        arr = np.moveaxis(arr[0, ...], 0, -1)
        img = sitk.GetImageFromArray(arr.astype(arr.dtype.newbyteorder("=")), isVector=True)
    else:

        if isinstance(channel, int):
            channel_number = channel
        else:
            if "omero" not in zarr_group.attrs or "channels" not in zarr_group.attrs["omero"]:
                raise ValueError("No OMERO metadata. Only integer channel numbers are supported.")
            omero_channel_labels = [c["label"] for c in zarr_group.attrs["omero"]["channels"]]

            if channel not in omero_channel_labels:
                raise Exception(f'Channel name "{channel}" is not in OMERO marker list: {omero_channel_labels}.')

            channel_number = omero_channel_labels.index(channel)

        if channel_number >= arr.shape[1] or channel_number < 0:
            raise ValueError("Channel number is out of range.")

        arr = arr[0, channel_number, ...]
        img = sitk.GetImageFromArray(arr.astype(arr.dtype.newbyteorder("=")), isVector=False)

    # Select XYZ spacing from input TCZYX
    img.SetSpacing(spacing[:1:-1])

    return img


def im_read_channel(filename, channel=None):  # noqa: C901
    """
    Read a channel from an image file.

    SimpleITK is used to read a channel from the file. Channel names can be used if the image file contains meta-data
    which can be parsed and provides names for the channels. Otherwise channel index should be used.

    :param filename: The path to an image file.
    :param channel: An integer for the channel index or string for the name of the channel. If None then all channel are
    read.

    :return: A SimpleITK Image.

    """

    filename = Path(filename)
    if filename.is_dir() and (filename / ".zattrs").exists():
        if _has_zarr:
            return _zarr_read_channel(filename, channel)
        else:
            raise ImportError("zarr is not installed.")

    reader = sitk.ImageFileReader()
    reader.SetFileName(str(filename))

    if channel is None:
        _logger.info('Reading whole "{}" image file.'.format(filename))
        return reader.Execute()

    ext = os.path.splitext(filename)[1].lower()
    if ext in [".tif", ".tiff"]:
        # OME TIFF files might be reader by the SCIFIOIMageIO, but that IO does not
        # share the OME XML in the meta data
        reader.SetImageIO("TIFFImageIO")

    reader.ReadImageInformation()

    _logger.debug(reader)

    if isinstance(channel, int):
        channel_number = channel

    else:
        IMAGE_DESCRIPTION = "ImageDescription"
        IMARIS_CHANNEL_INFORMATION = "imaris_channels_information"

        if IMAGE_DESCRIPTION in reader.GetMetaDataKeys():
            _logger.info('Found "{}" metadata field in {}.'.format(IMAGE_DESCRIPTION, filename))

            image_description = reader.GetMetaData(IMAGE_DESCRIPTION)
            _logger.debug(image_description)
            ome_info = OMEInfo(image_description)
            channel_names = ome_info.channel_names

        elif IMARIS_CHANNEL_INFORMATION in reader.GetMetaDataKeys():
            _logger.info('Found "{}" metadata field in {}.'.format(IMARIS_CHANNEL_INFORMATION, filename))
            imaris_data = reader.GetMetaData(IMARIS_CHANNEL_INFORMATION)
            _logger.debug(imaris_data)
            imaris_info = XMLInfo(imaris_data)

            channel_names = imaris_info.channel_names
        else:
            raise Exception(
                "No metadata information for channel names in file: {},"
                "reference channel by number not name.".format(filename)
            )

        if channel not in channel_names:
            raise Exception('Channel name "{0}" is not in filename marker list: {1}'.format(channel, channel_names))
        channel_number = channel_names.index(channel)
        _logger.debug('Channel name "{}" is channel number {}.'.format(channel, channel_number))

    if reader.GetImageIO() != "":
        reader.SetImageIO("")

    if reader.GetDimension() == 3:
        dimension_order = "XYZ"
    elif reader.GetDimension() == 4:
        dimension_order = "XYZC"
    elif reader.GetDimension() == 5:
        dimension_order = "XYZTC"
    else:
        raise Exception('File "{}" has unsupported dimension {}.'.format(filename, reader.GetDimension()))

    _logger.debug("dimension order: {} {}".format(dimension_order, reader.GetSize()))

    if reader.GetDimension() > 3 and reader.GetNumberOfComponents() > 1:
        _logger.warning(
            'File "{}" has {} dimensions and {} components. '
            "Ambiguous which contains the channels.".format(
                filename, reader.GetDimension(), reader.GetNumberOfComponents()
            )
        )

    # reduce dimensions
    extract_size = list(reader.GetSize())
    extract_index = [0] * len(extract_size)
    if reader.GetNumberOfComponents() == 1 and channel_number is not None:
        c_idx = dimension_order.upper().index("C")
        extract_index[c_idx] = channel_number

        if extract_index[c_idx] >= extract_size[c_idx]:
            raise Exception(
                "Channel index {} is dimension {} size {}.".format(channel_number, reader.GetDimension(), extract_size)
            )
        extract_size[c_idx] = 0

    # collapse time dimension
    if "T" in dimension_order:
        t_idx = dimension_order.upper().index("T")
        _logger.debug(
            'Dimension "T" ({}) has size {}, is being collapsed one first image.'.format(t_idx, extract_index[t_idx])
        )
        extract_size[t_idx] = 0

    reader.SetExtractSize(extract_size)
    reader.SetExtractIndex(extract_index)
    img = reader.Execute()

    if reader.GetNumberOfComponents() > 1 and channel_number is not None:
        if channel_number >= reader.GetNumberOfComponents():
            raise Exception(
                "Channel index {} is beyond the number of components {}".format(
                    channel_number, reader.GetNumberOfComponents()
                )
            )
        img = sitk.VectorIndexSelectionCast(img, channel_number)

    return img
